const config = require("config");
const Crypto = require("./Crypto");
const dayjs = require("dayjs");
const path = require("path");

class Order {
  static get STORAGE_MUTEX_KEY_PREFIX() {
    return "order_mutex:";
  }

  static get STORAGE_KEY_PREFIX() {
    return "order:";
  }

  static get PURCHASE_TAX_AMOUNT() {
    return 0.07;
  }

  // How long is a link between order <=> key stored
  static get PURCHASE_LINK_TIME_DAYS() {
    return 31;
  }
  static get PURCHASE_LINK_ORDER_TO_KEY_PREFIX() {
    return "order_link_order_key:";
  }
  static get PURCHASE_LINK_KEY_TO_ORDERS_PREFIX() {
    return "order_link_key_order:";
  }

  // How many minutes is a user allowed to take for finishing the payment
  static get PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS() {
    return 6;
  }

  static GET_ORDER_FILE_BASE_PATH(order_date) {
    let order_path = config.get("storage.data_path");
    if (process.env.NODE_ENV === "development") {
      order_path = path.join(order_path, "development");
    }
    order_path = path.join(
      order_path,
      order_date.format("YYYY"),
      order_date.format("MM")
    );

    return order_path;
  }
  /**
   * Data stored in Redis Database
   */
  #order_id;
  #expires_at;
  #amount;
  #price;
  #payment_completed;
  #payment_method_link; // Stores a link to an entry of the payment methods payment i.e. PayPal order id

  /**
   * Data populated by context
   */
  #order_date;
  #order_path;
  #redis_client;

  constructor(order_id, amount, price, payment_method_link, payment_completed) {
    this.#order_id = order_id;
    this.#expires_at = dayjs().add(6, "month");
    this.#order_date = dayjs.unix(this.#order_id.substr(0, 10));

    this.#order_path = Order.GET_ORDER_FILE_BASE_PATH(this.#order_date);

    this.#amount = parseInt(amount);
    this.#price = parseInt(price);

    if (payment_method_link) {
      this.#payment_method_link = payment_method_link;
    }

    if (payment_completed) {
      this.#payment_completed = payment_completed;
    }

    let Redis = require("ioredis");
    this.#redis_client = new Redis({
      host: config.get("redis.host"),
    });
  }

  getOrderID() {
    return this.#order_id;
  }

  getAmount() {
    return this.#amount;
  }

  getNettoPrice() {
    let amount = this.getPrice();
    let vat_amount = this.getVatAmount();
    return amount - vat_amount;
  }
  getVat() {
    return config.get("price.vat");
  }
  getVatAmount() {
    let vat = config.get("price.vat");
    let amount = this.getPrice();
    return Number((amount * (vat / 100)).toFixed(2));
  }
  getPrice() {
    return Number(this.#price).toFixed(2);
  }

  getPaymentMethodLink() {
    return this.#payment_method_link;
  }

  async setPaymentCompleted(completed) {
    return this.getWriteLock()
      .then(() => {
        this.#payment_completed = completed;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  isPaymentComplete() {
    return this.#payment_completed;
  }

  async setPaymentMethodLink(payment_method_link) {
    return this.getWriteLock()
      .then(() => {
        this.#payment_method_link = payment_method_link;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  static async CREATE_NEW_ORDER(amount, price, key) {
    return Order.GENERATE_UNIQUE_ORDER_ID().then(async (order_id) => {
      let new_order = new Order(order_id, amount, price);
      return new_order
        .getWriteLock()
        .then(() => new_order.createOrderLink(key))
        .then(() => new_order.save())
        .then(() => new_order.releaseWriteLock())
        .then(() => {
          return new_order;
        });
    });
  }

  static async LOAD_ORDER_FROM_ID(order_id) {
    let Redis = require("ioredis");
    let redis_client = new Redis({
      host: config.get("redis.host"),
    });
    return new Promise((resolve, reject) => {
      redis_client
        .hgetall(Order.STORAGE_KEY_PREFIX + order_id)
        .then((order_data) => {
          if (Object.keys(order_data).length === 0) {
            // Checking FS for order
            let order_date = dayjs.unix(order_id.substr(0, 10));
            let order_file = path.join(
              Order.GET_ORDER_FILE_BASE_PATH(order_date),
              order_id.toString() + ".json"
            );
            let fs = require("fs");
            if (fs.existsSync(order_file)) {
              order_data = JSON.parse(fs.readFileSync(order_file));
            } else {
              throw "Could not find Order in our database! Checking FS";
            }
          }
          let loaded_order = new Order(
            order_data.order_id,
            order_data.amount,
            order_data.price,
            JSON.parse(order_data.payment_method_link),
            order_data.payment_completed ? true : false
          );
          resolve(loaded_order);
        })
        .catch((reason) => {
          reject(reason);
        });
    });
  }

  async save() {
    let stored_data = {
      order_id: this.#order_id,
      expires_at: this.#expires_at.format("YYYY-MM-DD"),
      amount: this.#amount,
      price: this.#price,
      payment_completed: this.#payment_completed,
      payment_method_link: JSON.stringify(this.#payment_method_link),
    };
    /**
     * Completed Orders will be stored in Filesystem
     * Uncompleted Orders will be stored in Redis
     */
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (this.#payment_completed) {
      let fs = require("fs");
      let order_file = path.join(
        this.#order_path,
        this.#order_id.toString() + ".json"
      );
      // Create directory if it does not exist
      if (!fs.existsSync(path.dirname(order_file))) {
        fs.mkdirSync(path.dirname(order_file), { recursive: true });
      }
      let redis_client = this.#redis_client;
      return this.#redis_client
        .del(redis_key)
        .then(() => this.updateOrderLinkTTL())
        .then(() =>
          fs.writeFileSync(order_file, JSON.stringify(stored_data, null, 4))
        );
    } else {
      // Store Order in Redis
      let expiration = new dayjs();
      expiration = expiration.add(
        Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
        "hour"
      );
      return this.#redis_client
        .pipeline()
        .hmset(redis_key, stored_data)
        .expireat(redis_key, expiration.unix())
        .exec()
        .then(() => this.updateOrderLinkTTL());
    }
  }

  async delete() {
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (!this.isPaymentComplete()) {
      return this.deleteOrderLink()
        .then(() => this.#redis_client.del(redis_key))
        .then((deleted_keys) => {
          if (deleted_keys > 0) {
            return true;
          } else {
            return false;
          }
        })
        .catch(() => {
          return false;
        });
    } else {
      return false;
    }
  }

  async getWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;

    return new Promise(async (resolve, reject) => {
      let timestart = dayjs();
      let write_lock = 0;
      do {
        write_lock = await this.#redis_client
          .pipeline()
          .setnx(write_lock_key, 1)
          .expiretime(write_lock_key)
          .expire(write_lock_key, 15)
          .exec();
        let expire_at = write_lock[1][1];
        write_lock = write_lock[0][1];
        if (write_lock === 1) {
          resolve(true);
          break;
        } else if (dayjs().diff(timestart, "second") >= 15) {
          if (expire_at >= 0) {
            await this.#redis_client.expireat(write_lock_key, expire_at);
          }
          reject("Timed out waiting for write lock for Order");
          break;
        } else {
          if (expire_at >= 0) {
            await this.#redis_client.expireat(write_lock_key, expire_at);
          }
          await new Promise((resolve) => setTimeout(resolve, 1000));
        }
      } while (true);
    });
  }

  async releaseWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;
    return this.#redis_client.del(write_lock_key);
  }

  /**
   * Creates a link between Order and Key that will automatically expire
   * @param {string} key
   */
  async createOrderLink(key) {
    let expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    let redis_key_key_order = Order.PURCHASE_LINK_KEY_TO_ORDERS_PREFIX + key;
    return this.#redis_client
      .pipeline()
      .set(redis_key_order_key, key)
      .expireat(redis_key_order_key, expiration.unix())
      .rpush(redis_key_key_order, this.#order_id)
      .expireat(redis_key_key_order, expiration.unix())
      .exec();
  }

  async deleteOrderLink() {
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return this.getKeyFromOrderLink().then((key) => {
      let redis_key_key_order = Order.PURCHASE_LINK_KEY_TO_ORDERS_PREFIX + key;
      return this.#redis_client
        .pipeline()
        .del(redis_key_order_key)
        .lrem(redis_key_key_order, 0, this.#order_id)
        .exec();
    });
  }

  async updateOrderLinkTTL() {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;

    let expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
    if (!this.isPaymentComplete()) {
      expiration = dayjs().add(
        Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
        "hour"
      );
    }

    return this.#redis_client.expireat(redis_key, expiration.unix());
  }

  /**
   * Tries to get a key for this order. If the order link already expired this function rejects the promise
   */
  async getKeyFromOrderLink() {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return this.#redis_client.get(redis_key).then((result) => {
      if (result) {
        return result;
      } else {
        throw "Order Link does not exist";
      }
    });
  }

  static async GENERATE_UNIQUE_ORDER_ID() {
    // Generate Order ID => time in seconds since 1.1.1970 and and add a mutex to it to allow multiple order ids per second
    let Redis = require("ioredis");
    let redis_connection = new Redis({
      host: config.get("redis.host"),
    });
    let order_id = null;
    let order_base = Math.round(new Date().getTime() / 1000);
    let order_mutex = Math.floor(Math.random() * 10000);
    do {
      // make sure this order_id is not already registered
      order_id = "" + order_base + order_mutex;
      let order_lock = await redis_connection.setnx(order_id, true);
      if (order_lock === 1) {
        await redis_connection.expire(order_id, 5);
      } else {
        console.log("Couldn't acquire lock");
        order_mutex++;
      }
    } while (order_id === null);
    return order_id;
  }
}

module.exports = Order;
