var express = require("express");
var router = express.Router({ mergeParams: true });

const config = require("config");
const Order = require("../../app/Order.js");
const Key = require("../../app/Key.js");

const CLIENT_ID = config.get(`payments.paypal.client_id`);
const APP_SECRET = config.get(`payments.paypal.secret`);
const base = config.get(`payments.paypal.base`);
const webhook_redis_key = "checkout_paypal_webhook_id";

router.use("/", async (req, res, next) => {
  if (req.data && req.data.checkout) {
    req.data.checkout.payment = {
      provider: "paypal",
      paypal: {
        client_id: config.get(`payments.paypal.client_id`),
      },
    };
  }
  verifyWebhook()
    .then(() => {
      next("route");
    })
    .catch((reason) => {
      console.error(reason);
      next("route");
    });
});

router.get("/:funding_source", async (req, res) => {
  req.data.checkout.payment.paypal.funding_source = req.params.funding_source;

  if (req.params.funding_source === "card") {
    let client_token = await generateClientToken();
    req.data.checkout.payment.paypal.client_token = client_token;
  }

  req.data.change_url.funding_source =
    "/key/" +
    encodeURIComponent(req.data.key.key) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "#payment";

  req.data.change_url.funding_source_not_eligible =
    "/key/" +
    encodeURIComponent(req.data.key.key) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "?error=funding_source_not_eligible";

  req.data.change_url.order_base_url =
    "/key/" +
    encodeURIComponent(req.data.key.key) +
    "/checkout/" +
    encodeURIComponent(req.data.checkout.amount) +
    "/paypal/" +
    encodeURIComponent(req.data.checkout.payment.paypal.funding_source) +
    "/order";

  req.data.js.push("/js/checkout_paypal.js");
  res.render("key", req.data);
});

router.post("/:funding_source/order/create", async (req, res) => {
  // Order data is validated: Create and store the order in the redis database
  Order.CREATE_NEW_ORDER(
    // Create Order on our side
    req.params.amount,
    (req.params.amount / 300) * config.get("price.per_300"),
    req.data.key.key
  )
    .then(/** @param {Order} order */ (order) => createOrder(order)) // Create Order on PayPal Server
    .then((order_data) => {
      res.status(200).json(order_data);
    })
    .catch((reason) => {
      console.error(reason);
      res.status(400).json({
        errors: [{ msg: "Failed to create a new Order. Try again later" }],
      });
    });
});

router.post("/:funding_source/order/cancel", async (req, res) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id)
    .then((order) => {
      console.log("Loaded order");
      if (order.isPaymentComplete()) {
        // Not so good. Something went wrong after we captured the Payment
        // Refund it back
        let paypal_order_data = order.getPaymentMethodLink();
        if (paypal_order_data.name === "paypal") {
          console.log(paypal_order_data);
        }
      }
      return order.delete();
    })
    .then((success) => {
      if (success) {
        res.status(200).json({ msg: "Order deleted" });
      } else {
        res.status(400).json({ msg: "Order could not be deleted" });
      }
    })
    .catch((reason) => {
      console.error(reason);
      res.status(400).json({ msg: "Failed to load/cancel Order." });
    }); // Deletes a order but only if the payment is not yet completed
});

// We need to make sure that a card payment was authorized with 3-DS.
// All other PayPal captures will be handled in the next capture rotue
router.use("/card/order/capture", async (req, res, next) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id).then(
    /**
     * @param {Order} loaded_order
     */

    (loaded_order) => {
      let payment_method_link = loaded_order.getPaymentMethodLink();
      getOrderDetails(payment_method_link.order_id)
        .then((order_details) => {
          // A successful 3D Authentication by the user will result in liability_shift: POSSIBLE for the authentication result
          // Make sure the parameter is set for the order:
          if (
            order_details.payment_source.card.authentication_result
              .liability_shift !== "POSSIBLE"
          ) {
            console.log(JSON.stringify(order_details));
            throw "3D Authentication was not successful";
          } else {
            next("route");
          }
        })
        .catch((reason) => {
          console.error(reason);
          res.status(400).json({
            errors: [{ type: "PAYPAL_CARD_3D_ERROR" }],
          });
        });
    }
  );
});

// capture payment & store order information or fullfill order
router.post("/:funding_source/order/capture", async (req, res) => {
  //res.status(200).json({ test: "test" });
  //return;
  Order.LOAD_ORDER_FROM_ID(req.body.order_id).then(
    /**
     * @param {Order} loaded_order
     */
    (loaded_order) => {
      capturePayment(loaded_order)
        .then(() => loaded_order.setPaymentCompleted(true))
        .then(() => Key.CHARGE_KEY(req.data.key.key, loaded_order.getAmount()))
        .then(() => {
          let redirect_url =
            "/key/" + req.data.key.key + "/orders/" + loaded_order.getOrderID();
          res.status(200).json({
            redirect_url: redirect_url,
            order: {
              id: loaded_order.getOrderID(),
              amount: loaded_order.getAmount(),
              price: loaded_order.getPrice(),
            },
          });
        })
        .catch((reason) => {
          if (reason === "PAYMENT_NOT_COMPLETED_ERROR") {
            res.status(400).json({ errors: [{ type: reason }] });
          } else {
            res
              .status(400)
              .json({ errors: [{ msg: "Couldn't capture the payment" }] });
          }
        });
    }
  );
});

router.post("/webhook", async (req, res) => {
  // Verify that the webhook came from paypal
  let accessToken = await generateAccessToken();

  let verification_url = `${base}/v1/notifications/verify-webhook-signature`;

  fetch(verification_url, {
    method: "post",
    headers: {
      Authorization: `Bearer ${accessToken}`,
      "Content-Type": "application/json",
    },
    body: JSON.stringify({
      auth_algo: req.headers["paypal-auth-algo"],
      cert_url: req.headers["paypal-cert-url"],
      transmission_id: req.headers["paypal-transmission-id"],
      transmission_sig: req.headers["paypal-transmission-sig"],
      transmission_time: req.headers["paypal-transmission-time"],
      webhook_event: req.body,
      webhook_id: await getWebhookId(),
    }),
  })
    .then((response) => {
      if (response.status !== 200) {
        throw "Received status code " + response.status + " from PayPal API.";
      } else {
        return response.json();
      }
    })
    .then((response) => {
      if (
        !response.verification_status ||
        response.verification_status !== "SUCCESS"
      ) {
        console.error(response);
        throw "Webhook Verification was not successfull";
      } else {
        console.log(req.body.event_type);
        if (req.body.event_type === "PAYMENT.CAPTURE.COMPLETED") {
          console.log(JSON.stringify(req.body));
          // Check for a completed payment that did not get processed by us
          let order_id = req.body.resource.invoice_id;
          if (!order_id) {
            throw "No Order ID attached";
          } else {
            order_id = order_id.replace(/^INV_/, "");
          }
          return Order.LOAD_ORDER_FROM_ID(order_id).then(
            /** @param {Order} order */ (order) => {
              if (!order.isPaymentComplete()) {
                return order
                  .setPaymentCompleted(true)
                  .then(() => {
                    // Update Order status
                    let payment_link = order.getPaymentMethodLink();
                    payment_link.payment_status = req.body.resource.status;
                    return order.setPaymentMethodLink(payment_link);
                  })
                  .then(() => order.getKeyFromOrderLink())
                  .then((key) => Key.CHARGE_KEY(key, order.getAmount()));
              } else {
                throw "Order is already completed";
              }
            }
          );
        } else if (req.body.event_type === "PAYMENT.CAPTURE.REFUNDED") {
          console.log(JSON.stringify(req.body));
          // A Payment was refunded => Discharge the key
          let order_id = req.body.resource.invoice_id;
          if (!order_id) {
            throw "No Order ID attached";
          } else {
            order_id = order_id.replace(/^INV_/, "");
          }
          return Order.LOAD_ORDER_FROM_ID(order_id).then(
            /** @param {Order} order */ (order) => {
              if (order.isPaymentComplete()) {
                // Update Payment status
                let payment_link = order.getPaymentMethodLink();
                payment_link.payment_status = "REFUNDED";
                return order
                  .setPaymentMethodLink(payment_link)
                  .then(() => order.getKeyFromOrderLink())
                  .then((key) => Key.DISCHARGE_KEY(key, order.getAmount()));
              } else {
                throw "Order is already completed";
              }
            }
          );
        } else if (req.body.event_type === "CHECKOUT.ORDER.APPROVED") {
          let order_id = req.body.resource.purchase_units[0].invoice_id;
          if (!order_id) {
            throw "No Order ID attached";
          } else {
            order_id = order_id.replace(/^INV_/, "");
          }
          return Order.LOAD_ORDER_FROM_ID(order_id).then(
            /** @param {Order} order */ (order) => {
              let payment_method_link = order.getPaymentMethodLink();
              if (
                payment_method_link &&
                payment_method_link.payment_status &&
                payment_method_link.payment_status !== "COMPLETED"
              ) {
                return capturePayment(order);
              } else {
                throw "Order already captured";
              }
            }
          );
        } else {
          console.log(req.body);
          throw "Webhook not implemented";
        }
      }
    })
    .then(() => {
      res.status(200).send("");
    })
    .catch((reason) => {
      console.error(reason);
      res.status(200).json({ errors: [{ msg: "Error verifying Webhook" }] });
    });
});

module.exports = router;

//////////////////////

// PayPal API helpers

//////////////////////

// use the orders api to create an order

async function createOrder(loaded_order) {
  const accessToken = await generateAccessToken();

  const url = `${base}/v2/checkout/orders`;

  let unit_count = loaded_order.getAmount() / 300;
  let price_per_unit = loaded_order.getPrice() / unit_count;
  let tax_amount_per_unit = fixRounding(
    price_per_unit * Order.PURCHASE_TAX_AMOUNT,
    2
  );
  let item_amount_per_unit = fixRounding(
    price_per_unit - tax_amount_per_unit,
    2
  );

  let order = {
    intent: "CAPTURE",
    purchase_units: [
      {
        invoice_id: "INV_" + loaded_order.getOrderID(),
        description: "MetaGer Pass Einkauf",
        amount: {
          currency_code: "EUR",
          value: fixRounding(
            item_amount_per_unit * unit_count +
              tax_amount_per_unit * unit_count,
            2
          ),
          breakdown: {
            item_total: {
              currency_code: "EUR",
              value: fixRounding(item_amount_per_unit * unit_count, 2),
            },
            tax_total: {
              currency_code: "EUR",
              value: fixRounding(tax_amount_per_unit * unit_count, 2),
            },
          },
        },
        items: [
          {
            name: "MetaGer Schlüssel: Suchanfragen (300x)",
            quantity: unit_count,
            unit_amount: {
              currency_code: "EUR",
              value: item_amount_per_unit,
            },
            tax: {
              currency_code: "EUR",
              value: tax_amount_per_unit,
            },
            category: "DIGITAL_GOODS",
            description:
              "MetaGer Schlüssel zur werbefreien Nutzung der Suchmaschine MetaGer",
          },
        ],
      },
    ],
    application_context: {
      brand_name: "SUMA-EV",
      shipping_preference: "NO_SHIPPING",
    },
  };

  return fetch(url, {
    method: "post",
    headers: {
      "Content-Type": "application/json",
      Authorization: `Bearer ${accessToken}`,
    },
    body: JSON.stringify(order),
  })
    .then((response) => response.json())
    .then((data) => {
      if (data.status !== "CREATED") {
        console.error(data);
        throw "Couldn't create Payment";
      }
      return loaded_order
        .setPaymentMethodLink({
          name: "paypal",
          order_id: data.id,
          payment_status: data.status,
        })
        .then(() => {
          return { id: data.id, order_id: loaded_order.getOrderID() };
        });
    });
}

// use the orders api to capture payment for an order

/**
 *
 * @param {Order} order
 * @returns
 */
async function capturePayment(order) {
  const accessToken = await generateAccessToken();

  let paypal_order_id = order.getPaymentMethodLink().order_id;

  const url = `${base}/v2/checkout/orders/${paypal_order_id}/capture`;

  return fetch(url, {
    method: "post",
    headers: {
      "Content-Type": "application/json",
      Authorization: `Bearer ${accessToken}`,
    },
  })
    .then((response) => response.json())
    .then((response_data) => {
      console.log(JSON.stringify(response_data));
      if (
        response_data.status !== "COMPLETED" ||
        response_data.purchase_units[0].payments.captures[0].status !==
          "COMPLETED"
      ) {
        console.error(JSON.stringify(response_data));
        throw "PAYMENT_NOT_COMPLETED_ERROR";
      }
      return order.setPaymentMethodLink({
        name: "paypal",
        order_id: response_data.id,
        payment_status:
          response_data.purchase_units[0].payments.captures[0].status,
      });
    });
}

async function getOrderDetails(order_id) {
  let url = `${base}/v2/checkout/orders/${order_id}`;
  return generateAccessToken()
    .then((access_token) =>
      fetch(url, {
        headers: {
          Authorization: `Bearer ${access_token}`,
        },
      })
    )
    .then((response) => {
      if (response.status == 200) {
        return response.json();
      } else {
        throw "Couldn't retrieve Order";
      }
    });
}

async function getWebhookId() {
  let Redis = require("ioredis");
  let redis_client = new Redis({
    host: config.get("redis.host"),
  });
  return redis_client.get(webhook_redis_key);
}

async function verifyWebhook() {
  let domain = config.get("app.url").replace(/\/+$/, "");

  if (!domain.startsWith("https://")) {
    // Do not attempt to register a webhook for localhost etc.
    return;
  }

  let Redis = require("ioredis");
  let redis_client = new Redis({
    host: config.get("redis.host"),
  });
  if (await redis_client.get(webhook_redis_key)) {
    return;
  }
  const accessToken = await generateAccessToken();
  return fetch(`${base}/v1/notifications/webhooks`, {
    method: "get",
    headers: {
      "Content-Type": "application/json",
      Authorization: `Bearer ${accessToken}`,
    },
  })
    .then((response) => response.json())
    .then((webhooks) => {
      let webhook_id = undefined;
      webhooks.webhooks.forEach(async (webhook) => {
        if (webhook.url.startsWith(domain)) {
          webhook_id = webhook.id;
        } else if (webhook.url.match(/^https:\/\/.*\.ngrok\.io/)) {
          // Another ngrok development webhook. Delete it.
          let promises = [];
          webhook.links.forEach((link) => {
            if (link.rel === "delete") {
              promises.push(
                fetch(link.href, {
                  method: link.method,
                  headers: {
                    Authorization: `Bearer ${accessToken}`,
                  },
                })
              );
            }
          });
          await Promise.all(promises);
        }
      });
      if (webhook_id !== undefined) {
        return redis_client
          .setex(webhook_redis_key, 3600, webhook_id)
          .then(() => {
            throw "WEBHOOK_ALREADY_REGISTERED"; // Webhook already registered
          });
      } else {
        // Webhook does not yet exist
        console.log("Webhook does not yet exist");
        return true;
      }
    })
    .then(async () => {
      // Webhook does not exist yet
      return fetch(`${base}/v1/notifications/webhooks`, {
        method: "post",
        headers: {
          "Content-Type": "application/json",
          Authorization: `Bearer ${accessToken}`,
        },
        body: JSON.stringify({
          event_types: [
            { name: "PAYMENT.CAPTURE.COMPLETED" },
            { name: "PAYMENT.CAPTURE.REFUNDED" },
            { name: "PAYMENT.CAPTURE.REVERSED" },
            { name: "CHECKOUT.ORDER.APPROVED" },
          ],
          url: domain + "/webhooks/paypal/webhook",
        }),
      })
        .then((response) => {
          if (response.status !== 201) {
            return Promise.reject(response);
          } else {
            return response.json();
          }
        })
        .then((response_data) => {
          console.log(JSON.stringify(response_data));
          redis_client.setex(webhook_redis_key, 3600, response_data.id);
        });
    })
    .catch((reason) => {
      if (reason !== "WEBHOOK_ALREADY_REGISTERED") {
        console.error("Could not register Webhook for PayPal.");
        console.error(reason);
      }
    });
}

async function refundPayment(capture_ids) {
  const accessToken = await generateAccessToken();
  let promises = [];
  capture_ids.forEach((capture_id) => {
    promises.push(
      fetch(`${base}/v2/payments/captures/${capture_id}/refund`, {
        method: "post",
        headers: {
          Authorization: `Bearer ${accessToken}`,
          "Content-Type": "application/json",
        },
        body: JSON.stringify({
          note_to_payer: "Something went wrong when processing your payment.",
        }),
      })
    );
  });
  return Promise.all(promises);
}

// Client Token for handling credit card payments
async function generateClientToken() {
  const accessToken = await generateAccessToken();

  const response = await fetch(`${base}/v1/identity/generate-token`, {
    method: "post",
    headers: {
      Authorization: `Bearer ${accessToken}`,
      "Accept-Language": "en_US",
      "Content-Type": "application/json",
    },
  });

  const data = await response.json();

  return data.client_token;
}

// generate an access token using client id and app secret

async function generateAccessToken() {
  const auth = Buffer.from(CLIENT_ID + ":" + APP_SECRET).toString("base64");

  const response = await fetch(`${base}/v1/oauth2/token`, {
    method: "post",

    body: "grant_type=client_credentials",

    headers: {
      Authorization: `Basic ${auth}`,
    },
  });

  const data = await response.json();

  return data.access_token;
}

/**
 *
 * Javascripts calculations with floating points produce weird artifacts. This method is to fix that issue
 *
 * @param {Float} value
 * @param {Int} precision
 * @returns
 */
function fixRounding(value, precision) {
  var power = Math.pow(10, precision || 0);
  return Math.round(value * power) / power;
}
